/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package javax.servlet.websockets;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.websockets.EventType.EventSubType;

import org.apache.catalina.comet.CometEvent;
import org.apache.catalina.comet.CometProcessor;
import org.apache.coyote.websockets.WebSocketsSessionImpl;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;
import org.jwebsocket.config.JWebSocketCommonConstants;
import org.jwebsocket.kit.RequestHeader;
import org.jwebsocket.kit.WebSocketHandshake;

/**
 * The base servlet which all WebSocket servlets need to implement.  Similar to an
 * HTTP Servlet
 * 
 * @author Ant Kutschera
 *
 * @param <T> the type of data object to receive
 */
public class WebSocketsServlet<T> extends HttpServlet implements CometProcessor {

	private static final long serialVersionUID = 1L;

	private static final String SEC_WEB_SOCKET_PROTOCOL = "Sec-WebSocket-Protocol";

	@Override
	public final void service(ServletRequest req, ServletResponse res)
			throws ServletException, IOException {
		//final, so business programmer cant override it
		super.service(req, res);
	}
	
	@Override
	protected final void service(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		//final, so business programmer cant override it
		
		//do the handshake
		
		request.getHeader("upgrade");

        String upgradeToWebSockets = request.getHeader("upgrade");
        if(upgradeToWebSockets != null && upgradeToWebSockets.equals("WebSocket")){

        	try{
				//messy - i would love to read these from the inputstream, but it didnt work!
        		byte[] payload = (byte[]) request.getAttribute("__websessions-body");
 
				//this bit doesnt work at all :-(
        		if(request.getContentLength() >= 0){
        			byte[] payload2 = new byte[request.getContentLength()];
        			int curr = -1;
        			InputStream is = request.getInputStream();
        			boolean foundMoreThanDefaults = false;
        			for(int i = 0; i < request.getContentLength(); i++){
        				curr = is.read();
        				if(curr != -1){
        					payload2[i] = (byte)curr;
        					if(curr != 0){
        						foundMoreThanDefaults = true;
        					}
        				}
        			}
        			
        			if(foundMoreThanDefaults){
        				payload = payload2;
        			}
        			//doesnt work either: request.getInputStream().read(payload, 0, request.getContentLength());
        		}

        		
        		//the damn input stream doesnt contain the payload we need to complete the handshake
        		//luckily, our websockets processor added it to the request as an attribute :-)
				byte[] reassembledRequest = WebSocketsHandshakeHelper.reassembleRequest(request, payload);
System.out.println("reassembled request: \r\n-----");
System.out.println(new String(reassembledRequest, "US-ASCII"));
System.out.println("-----\r\n");
				
				RequestHeader header = WebSocketsHandshakeHelper.handshake(reassembledRequest, response);
				
//				HTTP/1.1 101 WebSocket Protocol Handshake
//				Upgrade: WebSocket
//				Connection: Upgrade
//				Sec-WebSocket-Origin: http://localhost:8080
//				Sec-WebSocket-Location: ws://localhost:8085/ch.maxant.blog.nio.websocketshandshake
//
//				ê´64#▀Ù↓L▬♣¹·¾¨÷
	
				//host=localhost:8085
				//origin=http://localhost:8080
				//location=ws://localhost:8085/ch.maxant.blog.nio.websocketshandshake
				//path=/ch.maxant.blog.nio.websocketshandshake
				//searchString=
				//args={prot=json}
				//payload=[B@fa9557}
				
				response.setHeader("Upgrade", "WebSocket");
				response.setHeader("Connection", "Upgrade");
				response.setHeader("Sec-WebSocket-Origin", header.getString("origin"));
				String location = header.getString("location");
				response.setHeader("Sec-WebSocket-Location", location);
				
				if(request.getHeader(SEC_WEB_SOCKET_PROTOCOL) != null){
					response.setHeader(SEC_WEB_SOCKET_PROTOCOL, request.getHeader(SEC_WEB_SOCKET_PROTOCOL));
				}

				response.setContentLength(16); //so that the container doesnt use chunking!
				
				response.setStatus(HttpServletResponse.SC_SWITCHING_PROTOCOLS); // using i18n, the associated message is actually that as required in the websockets spec: "WebSocket Protocol Handshake"

				response.getOutputStream().write((byte[])(header.get("payload")));
				response.flushBuffer();

			} catch (IOException e) {
				log("unable to find session - did you send a handshake request?");
			}
        }else{
        	response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Expected a WebSockets Handshake request!");
        }
		
	}
	
	/**
	 * called by the container when a client closes.
	 * default impl does nothing.
	 * 
	 * @param eventSubType the reason for the close.
	 * @param session the session of the client being closed.  this session will be tidied up immediately after this call.
	 */
	public void doClose(EventType.EventSubType eventSubType, WebSocketsSession session){
	}

	/**
	 * called by the container when an error occurs.
	 * default impl does nothing.
	 * 
	 * @param eventSubType the type of error.
	 * @param session the session in which this error occurred.
	 * @param e the originating exepction, OR NULL.
	 */
	protected void doError(EventType.EventSubType eventSubType, WebSocketsSession session, Exception e){
	}

	/**
	 * called by the container when a websocket frame has been read.
	 * 
	 * @param frame the frame just read
	 * @param request the request related to the initial handshake
	 * @param response the response to write to, if required.
	 * @throws IOException if anything goes wrong
	 */
	private void doMessageInternal(
			byte[] frame, 
			WebSocketsServletRequest<T> request, 
			WebSocketsServletResponse response) throws IOException {

		request.setData(frame);
		doMessage(request, response);
	}
	
	/**
	 * called by the container when a websocket frame has been read.
	 * delegates to the websockets-protocol-relevant doXXXMessage method.
	 * feel free to override this if you like.
	 */
	protected void doMessage(WebSocketsServletRequest<T> request, WebSocketsServletResponse response) throws IOException {
		String protocol = request.getWebSocketsProtocol();
		if(protocol.equals(WebSocketsServletRequest.JSON)){
			doJSONMessage(request, response);
		}else if(protocol.equals(WebSocketsServletRequest.XML)){
			doXMLMessage(request, response);
		}else if(protocol.equals(WebSocketsServletRequest.TEXT)){
			doTextMessage(request, response);
		}else if(protocol.equals(WebSocketsServletRequest.BIN)){
			doBinaryMessage(request, response);
		}else{
			doError(EventType.EventSubType.IOEXCEPTION, request.getSession(), new IllegalArgumentException("Unknown protocol " + protocol));
		}
	}

	/**
	 * called by the default implementation of the {@link #doMessage(byte[], HttpSession, OutputStream)} method, 
	 * when the requested protocol is JSON.
	 * @param request the original request as related to the handshake.
	 * @param response the response used to send data back to the client or to other clients.
	 * @throws IOException should anything go wrong.
	 */
	protected void doJSONMessage(WebSocketsServletRequest<T> request, WebSocketsServletResponse response) throws IOException{};

	/**
	 * called by the default implementation of the {@link #doMessage(byte[], HttpSession, OutputStream)} method, 
	 * when the requested protocol is XML.
	 * @param request the original request as related to the handshake.
	 * @param response the response used to send data back to the client or to other clients.
	 * @throws IOException should anything go wrong.
	 */
	protected void doXMLMessage(WebSocketsServletRequest<T> request, WebSocketsServletResponse response) throws IOException{};
	
	/**
	 * called by the default implementation of the {@link #doMessage(byte[], HttpSession, OutputStream)} method, 
	 * when the requested protocol is TEXT.
	 * @param request the original request as related to the handshake.
	 * @param response the response used to send data back to the client or to other clients.
	 * @throws IOException should anything go wrong.
	 */
	protected void doTextMessage(WebSocketsServletRequest<T> request, WebSocketsServletResponse response) throws IOException{};
	
	/**
	 * called by the default implementation of the {@link #doMessage(byte[], HttpSession, OutputStream)} method, 
	 * when the requested protocol is BIN (binary).
	 * @param request the original request as related to the handshake.
	 * @param response the response used to send data back to the client or to other clients.
	 * @throws IOException should anything go wrong.
	 */
	protected void doBinaryMessage(WebSocketsServletRequest<T> request, WebSocketsServletResponse response) throws IOException{};
	
	/**
	 * called by the container when client initially connects.  there is no response passed here, because 
	 * WebSockets do not allow for data to be sent back to the client during this stage.
	 * 
	 * things like the requested URI can be examined.  the {@link WebSocketsSession}
	 * is also available from the request.
	 * 
	 * default impl does nothing.
	 * 
	 * @param request the request as related to the HTTP handshake request.  
	 */
	protected void doOpen(WebSocketsServletRequest<T> request){
	}

	/**
	 * handles the event from the container.  do not override this!
	 */
	@Override
	public final void event(CometEvent event) throws IOException, ServletException {
		if(event.getEventType().equals(CometEvent.EventType.BEGIN)){
			//get the session now, so that we get it before committing the response
			event.getHttpServletRequest().getSession(true); 

			service(event.getHttpServletRequest(), event.getHttpServletResponse());
			
			doOpen(event.getWebSocketsServletRequest());
		}else if(event.getEventType().equals(CometEvent.EventType.READ)){
			
			try{
				InputStream is = event.getHttpServletRequest().getInputStream();
				byte[] buf = new byte[1];
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				boolean inFrame = true;
				do {
					int n = is.read(buf); //can throw an IOException
					if (n > 0) {
						if(!inFrame){
							if(buf[0] == 0){
								inFrame = true;
							}else{
								//ignore this data?
							}
						}else{
							if(buf[0] == -1){
								//end of frame
								inFrame = false;
								byte[] data = baos.toByteArray();
								try{
									doMessageInternal(data, 
											event.getWebSocketsServletRequest(),
											event.getWebSocketsServletResponse());
								}catch(Exception e){
									log("Failed to call doMessage for " + new String(data), e);
									error(event, e);
								}
								baos.reset();
							}else if(buf[0] == 0){
								//ignore
							}else{
								//collect the data!
								baos.write(buf);
							}
						}
					} else if (n < 0) {
						error(event, null);
						return;
					}
				} while (is.available() > 0);	
				
				System.out.print("");
				
			}catch(IOException e){
				error(event, e);
			}
		}else if(event.getEventType().equals(CometEvent.EventType.ERROR)){
			error(event, null);
		}else if(event.getEventType().equals(CometEvent.EventType.END)){
			doClose(mapSubEvent(event.getEventSubType()), new WebSocketsSessionImpl(event.getHttpServletRequest().getSession()));
			event.close();
		}else {
			throw new IllegalArgumentException("unknown event type " + event.getEventType() + " - code change required");
		}
	}

	private EventSubType mapSubEvent(
			org.apache.catalina.comet.CometEvent.EventSubType eventSubType) {
		
		if(eventSubType == null){
			eventSubType = org.apache.catalina.comet.CometEvent.EventSubType.IOEXCEPTION;
		}
		
		switch(eventSubType){
		case IOEXCEPTION: return EventSubType.IOEXCEPTION;
		case CLIENT_DISCONNECT: return EventSubType.CLIENT_DISCONNECT;
		case SERVER_SHUTDOWN: return EventSubType.SERVER_SHUTDOWN;
		case SESSION_END: return EventSubType.SESSION_END;
		case TIMEOUT: return EventSubType.TIMEOUT;
		case WEBAPP_RELOAD: return EventSubType.WEBAPP_RELOAD;
		default: 
			return null;
		}
	}

	private void error(CometEvent event, Exception e) throws IOException {
		doError(mapSubEvent(event.getEventSubType()), new WebSocketsSessionImpl(event.getHttpServletRequest().getSession()), e);
		event.close();
	}

	/** partly taken from jWebSockets! lgpl - does that conflict with apache?  hope not.  this is only a proof of concept! */
	private static final class WebSocketsHandshakeHelper {

	    private static final Log log = LogFactory.getLog(WebSocketsHandshakeHelper.class);

		private WebSocketsHandshakeHelper(){}
		
		public static byte[] reassembleRequest(HttpServletRequest request, byte[] payload){

			//request string
			String r = request.getMethod() + " " + request.getRequestURI() + " " + request.getProtocol() + "\r\n";
			//TODO add the query string too?
			//request.queryString();

			//headers
			Enumeration<String> e = request.getHeaderNames();
			while(e.hasMoreElements()){
				String name = e.nextElement();
				name = getProperlyCamelizedName(name);
				r += name + ": " + request.getHeader(name) + "\r\n";
			}
			r += "\r\n"; //blank line to indicate start of body
			
			try {
				//charset, as required by WebSocketHandshake!
				byte[] top = r.getBytes("US-ASCII");

				byte[] result = new byte[top.length + payload.length];
				
				System.arraycopy(top, 0, result, 0, top.length);
				System.arraycopy(payload, 0, result, top.length, payload.length);
				
				return result;
			} catch (UnsupportedEncodingException ex) {
				//should never ever happen!
				throw new RuntimeException("cannot recreate request from request", ex);
			} 
		}

		private static String getProperlyCamelizedName(String name) {
			//stupid dumb assed tomcat is making all the names small cased!
//			upgrade = WebSocket
//			connection = Upgrade
//			host = localhost:8081
//			origin = http://localhost:8080
//			sec-websocket-key1 = 11 7K1   01}   22 4wI8
//			sec-websocket-key2 = I898|9 %5L\386 ]W1 ;

			if(name.equals("origin")){
				return "Origin";
			}else if(name.equals("connection")){
				return "Connection";
			}else if(name.equals("host")){
				return "Host";
			}else if(name.equals("upgrade")){
				return "Upgrade";
			}else if(name.equals("sec-websocket-key1")){
				return "Sec-WebSocket-Key1";
			}else if(name.equals("sec-websocket-key2")){
				return "Sec-WebSocket-Key2";
			}else if(name.equals("sec-webSocket")){
				return "Sec-WebSocket";
			}else{
				return name;
			}
		}

		public static RequestHeader handshake(byte[] input, HttpServletResponse response) throws IOException {
			@SuppressWarnings("rawtypes")
			Map lRespMap = WebSocketHandshake.parseC2SRequest(input);
			// maybe the request is a flash policy-file-request
			String lFlashBridgeReq = (String) lRespMap.get("policy-file-request");
			if (lFlashBridgeReq != null) {
				log.warn("TCPEngine returned policy file request ('"
						+ lFlashBridgeReq + "'), check for FlashBridge plug-in.");
			}
			// generate the websocket handshake
			// if policy-file-request is found answer it
			byte[] lBA = WebSocketHandshake.generateS2CResponse(lRespMap);
			if (lBA == null) {
				if (log.isDebugEnabled()) {
					log.warn("TCPEngine detected illegal handshake.");
				}
				return null;
			}
			//if (log.isDebugEnabled()) {
			//	log.debug("Handshake Response:\n" + new String(lResp));
			//	mLog.debug("Flushing initial WebSocket handshake...");
			//}
			
//			response.getWriter().write(new String(lBA, "US-ASCII"));

System.out.println("responded with\r\n-----\r\n" + new String(lBA) + "\r\n-----");
			
			// if we detected a flash policy-file-request return "null"
			// (no websocket header detected)
			if (lFlashBridgeReq != null) {
				log.warn("TCPEngine returned policy file response ('" + new String(lBA, "US-ASCII") + "'), check for FlashBridge plug-in.");
				return null;
			}

			RequestHeader lHeader = new RequestHeader();
			Map<String, String> lArgs = new HashMap<String, String>();
			String lPath = (String) lRespMap.get("path");

			// isolate search string
			String lSearchString = "";
			if (lPath != null) {
				int lPos = lPath.indexOf(JWebSocketCommonConstants.PATHARG_SEPARATOR);
				if (lPos >= 0) {
					lSearchString = lPath.substring(lPos + 1);
					if (lSearchString.length() > 0) {
						String[] lArgsArray = lSearchString.split(JWebSocketCommonConstants.ARGARG_SEPARATOR);
						for (int i = 0; i < lArgsArray.length; i++) {
							String[] lKeyValuePair = lArgsArray[i].split(JWebSocketCommonConstants.KEYVAL_SEPARATOR, 2);
							if (lKeyValuePair.length == 2) {
								lArgs.put(lKeyValuePair[0], lKeyValuePair[1]);
								if (log.isDebugEnabled()) {
									log.debug("arg" + i + ": " + lKeyValuePair[0] + "=" + lKeyValuePair[1]);
								}
							}
						}
					}
				}
			}

			if (log.isDebugEnabled()) {
				log.debug("Handshake flushed.");
			}

			lHeader.put("host", lRespMap.get("host"));
			lHeader.put("origin", lRespMap.get("origin"));
			lHeader.put("location", lRespMap.get("location"));

			lHeader.put("path", lRespMap.get("path"));
			lHeader.put("searchString", lSearchString);
			lHeader.put("args", lArgs);

			byte[] payload = new byte[16];
			for(int i = 0; i < 16; i++){
				payload[i] = lBA[lBA.length-16+i];
			}
			lHeader.put("payload", payload);

			lHeader.put("raw", lBA);
			
			return lHeader;
		}

	}

}
