/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package javax.servlet.websockets;

import java.io.IOException;

import javax.servlet.ServletResponse;

/**
 * Extends the {@link ServletResponse} interface to provide websocket-specific
 * functionality in sending data to clients.
 * <p>
 * 
 * @author Ant Kutschera
 * @version $Version$
 * @see javax.servlet.ServletResponse
 */
public interface WebSocketsServletResponse extends ServletResponse {

    /**
     * Encodes the specified URL by including the session ID in it, or, if
     * encoding is not needed, returns the URL unchanged. The implementation of
     * this method includes the logic to determine whether the session ID needs
     * to be encoded in the URL. For example, if the browser supports cookies,
     * or session tracking is turned off, URL encoding is unnecessary.
     * <p>
     * For robust session tracking, all URLs emitted by a servlet should be run
     * through this method. Otherwise, URL rewriting cannot be used with
     * browsers which do not support cookies.
     * 
     * @param url
     *            the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     *         otherwise.
     */
    public String encodeURL(String url);
    
    /**
     * sends the given data to the client as a JSON string.
     * encodes using UTF-8.
     * 
     * @param data the data to send.
     * @throws IOException 
     */
    public void sendJSONResponse(Object data) throws IOException;
    
    /**
     * sends the given data to the client as an XML string.
     * encodes using UTF-8.
     * 
     * @param data the data to send.
     * @throws IOException 
     */
    public void sendXMLResponse(Object data) throws IOException;
    
    /**
     * sends the given data to the client as a string.
     * encodes using UTF-8.
     * 
     * @param data the data to send.
     * @throws IOException 
     */
    public void sendTextResponse(String data) throws IOException;
    
    /**
     * sends the given data to the client as raw bytes.
     * encodes using UTF-8.
     * 
     * @param data  the data to marshal and send.
     * @throws IOException 
     */
    public void sendBinaryResponse(byte[] data) throws IOException;

    /**
     * called by the servlet to close the connection to the client.
     * results in the 
     * {@link WebSocketsServlet#doClose(javax.servlet.websockets.EventType.EventSubType, WebSocketsSession)} 
     * method being called too, so that the servlet can do any tidying up that is required.
     * @throws IOException 
     */
    public void close() throws IOException;

}
